@q vim:set ts=8 syntax=cweb: @>
\datethis
\def\em{\bf}
\font\ff=cmr12
\ff
This program converts Windows TrueType Fonts into Type1 font for use in \TeX.
It's originally writen by He Qiang\footnote\dag{\em obase@@163.net}. 
Modified by Wang Yin\footnote\ddag{\em wang-y01@@mails.tsinghua.edu.cn} 
into a \.{CWEB} file and changed some portions of the code.

@* The gbkfonts Program. 

The program is making use of several other program from the FreeType Project 
({\em http://www.freetype.org}) such as \.{ttf2pfb, t1asm, type1afm, afm2tfm}.
The following is the overall structure of the program text.

@c
@<Included files@>@/
@<Global variables@>@/
@<Utility functions@>@/
@<The |main| program@>@/

@ @<Include...@>=
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include "freetype.h"
#include "ftxpost.h"
#include "t1lib.h"

#ifndef WIN32
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#else
#include <windows.h>
#include  <io.h>


#define mkdir(a,b) _mkdir(a)
#define access _access

#endif

#define INSERT


char* LookUp_Name(int);

@ the variables pid and eid are set my ttf2pfb.c.
@d NAMELEN 128
@<Global...@>=
extern char fontname[], fam_name[];
unsigned short pid, eid;

@ Use the function |LookUp_Name()| defined in ttf2pfb.c
get eid and PostScript font name and family name if not specified.
@ @<get eid...@>=
if (!Init_Font_Engine(ttf_file))
	fatal(_("can't open TrueType engine\n"));

if (fontname[0] == 0) {
	strcpy (fontname, LookUp_Name(6));
	if (!bs) fprintf(stderr, _("** Detected font name : %s\n"), fontname);
}

if (fam_name[0] == 0) {
	strcpy(fam_name, LookUp_Name(1));
	if (!bs) fprintf(stderr, _("** Detected family name : %s\n"), fam_name);
}

@ Create fd files such as C19song.fd for CJK package. 
I omit the $\backslash$CJKbold option in the fd file, 
since most user don't use them.

@<create fd...@>= 
	/* write fd file for latex */
	sprintf(cmd, "tex/latex/CJK/GB/c19%s.fd", facename);
	fd_file = fopen (cmd, "w");
	fprintf(fd_file, "%% This is the file c19%s.fd of the CJK package\n", facename);
	fprintf(fd_file, "%% for using Asian logographs (Chinese/Japanese/Korean) with LaTeX2e\n");
	fprintf(fd_file, "%% created by Werner Lemberg <wl@@gnu.org>\n");
	fprintf(fd_file, "%% Version 4.3.0 (20-Jun-1999)\n\n");
	fprintf(fd_file, "\\def\\fileversion{4.3.0}\n");
	fprintf(fd_file, "\\def\\filedate{1999/06/20}\n\n");
	fprintf(fd_file, "%% \\ProvidesFile{c19%s.fd}[\\filedate\\space\\fileversion]\n", facename);
	fprintf(fd_file, "%% Chinese characters (extension of GB 2312)\n");
	fprintf(fd_file, "%% character set: GBK\n");
	fprintf(fd_file, "%% font encoding: CJK (extended)\n\n");
	fprintf(fd_file, "\\DeclareFontFamily{C19}{%s}{}\n", facename);
	fprintf(fd_file, "\\DeclareFontShape{C19}{%s}{m}{n}{<-> CJK * %s%s}{}\n",
			facename, prefix, facename);
	fprintf(fd_file, "\\DeclareFontShape{C19}{%s}{bx}{n}{<-> CJKb * %s%s}{}\n", 
			facename, prefix, facename);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{m}{it}{<-> CJK * %s%ssl}{}\n", 
		facename, prefix, facename);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{bx}{it}{<-> CJKb * %s%ssl}{}\n", 
		facename, prefix, facename);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{m}{sl}{<-> CJK * %s%ssl}{}\n",   
		facename, prefix, facename);
	fprintf(fd_file, 
		"\\DeclareFontShape{C19}{%s}{bx}{sl}{<-> CJKb * %s%ssl}{}\n", 
		facename, prefix, facename);
	fprintf(fd_file, "\n\\endinput\n");
	fclose(fd_file);


@ This function get the file basename's pointer.
@<Utility...@>=
char *trimpath(char* str)
{	
	
	char* ptr;
	ptr = strrchr(str, '\\');
	if (ptr == 0) ptr = strrchr(str, '/');
	if (ptr == 0) return str; else return ptr+1;
}

@ @<variables us...@>=
int  i, fontuid = 0, nn;
FILE *pfbmapfile, *ttfmapfile, *fout, *dvipdfm_mapfile, *fd_file;

static char *prefix = "gbk", *ttf_file, *ttf_basename, *facename;
static char slface[NAMELEN], name[NAMELEN], slname[NAMELEN], plane[8];
static char cmd[NAMELEN], cmd1[NAMELEN]; 

/* boolean switches */
static int  bs = 0, bv = 0, bf = 0, font_name_detected = 0;


@ set the names. for example, if you use\hfil\break
\tt gbkfonts -c -gbk c:/winnt/fonts/simsun.ttf song \rm

you get the names: name = ``gbksong??'' , slname = ``gbksongsl??''

@<set appro...@>=
sprintf(plane, "%02d", nn);
sprintf(name,  "%s%s%s", prefix, facename, plane);
sprintf(slname,"%s%s%s", prefix, slface, plane);

@ Write configuration files for several programs.

The {\tt dvipdfmx} program from the dvipdfm-cjk project use cid-x.map to 
find TTF Chinese fonts. 

Pdftex use pdftex.cfg to find the map files. The map files direct it to the TTF
fonts. They are the same as used by {\tt dvips} when {\tt dvips} use PK fonts.

The {\tt CJK} package find it's font file name from the .fd files.

@<write config...@>=
	/* write cid-x for dvipdfm */
	fout = fopen ("cid-x.map", "a");
	fprintf(fout, "%s%s@@UGBK@@\tUniGB-UCS2-H\t:0:%s\n", 
			prefix, facename, ttf_basename);

	fprintf(fout, "%s%ssl@@UGBK@@\tUniGB-UCS2-H\t:0:%s\n", 
			prefix, facename, ttf_basename);
	fclose (fout);

	/* write pdftex.cfg */
	fout = fopen ("pdftex.cfg", "a");
	fprintf(fout, "map +%s%s.map\n", prefix, facename);
	fclose (fout);

	/* write map files for dvips */
	fout = fopen ("ttfonts.map", "a");
	fprintf(fout, "%s%s@@UGBK@@    %s  Pid = 3 Eid = %d\n", prefix, facename, ttf_basename, eid);
	fprintf(fout, "%s%s@@UGBK@@    %s  Slant = 0.167 Pid = 3 Eid = %d\n", 
			prefix, slface, ttf_basename, eid);
	fclose (fout);

	@<create fd file for LaTeX CJK@>@/

@ @<Utility...@>=
void showhelp()
{
	puts(_("Usage:  gbkfonts [-f] [-name str] [-prefix str] ttf_file facename"));
	puts(_("\t          -f: fast, only create config and map files"));
	puts(_("\t        -uid: select an unique ID for your font"));
	puts(_("\t       -name: select an PostScript font name for you font"));
	puts(_("\t -prefix str: add the prefix rather than \"gbk\" to you font name"));
	puts(_("\t    ttf_file: TTF file name"));
	puts(_("\t    facename: song/hei/kai/li/fs, etc."));
	puts("");
	puts(_("Example: gbkfonts -c -gbk /home/knuth/fonts/simsun.ttf song"));
	puts("");
	puts(_("Then copy folder \"dvips\", \"tex\", \"dvipdfm\", and \"fonts\" to TEXMFLOCAL"));
	puts(_("Append file cjk.map, cid-x.map, pdftex.cfg, ttfonts.map to your system"));
	puts(_("The script \"appendconf\" can help you append these files to the system"));
	puts(_("The script \"go\" can help you create all fonts in a directory. Modify it as you like."));
}

@ @<The |main| progra...@>=
int main(int argc, char* argv[])
{
	@<variables used in |main| @>;
#ifdef INTERNATIONAL
	@<set locale information and register to gettext@>;
#endif
	@<parse arguments, test font file existence@>;
	@<get eid and fontname infomation@>;
	@<create directory to store fonts and other files@>;
	@<open map files for write@>;

	if (!bs) fprintf(stderr, _("** Creating %s%s from %s\n"), prefix, facename, ttf_file);@/

	for (nn = 1; nn <= 94; nn++) {
		
		if (!bs) fprintf(stderr, "[%d]", nn); 
		@<set appropriate names {\tt name}, {\sl slname} $\ldots$ for the font@>;

		if (bf) goto skipfontmake;
		@<make the fonts: {\sl pfb, afm, tfm} @>;
skipfontmake:
		@<add line to map files@>;
	}

	fputs("\n", stderr); 
	@<close mapfiles@>;
	@<write config files@>;
	return 0;
}

@ @<Include...@>=
#ifdef INTERNATIONAL 

#include <libintl.h>
#include <locale.h>
#define _(String) gettext (String)
#define PACKAGE "gbkfonts"
#define LOCALEDIR "/usr/share/locale"

#else 
#define _(String) String

#endif

@ @<varia...@>=
char *lc_all;

@ @<set lo...@>=
lc_all = setlocale(LC_ALL, "");
if (! bindtextdomain (PACKAGE, NULL))
		fprintf(stderr, "Can't bind to locale %s\n", lc_all);

puts(textdomain (PACKAGE));
if (textdomain (PACKAGE) == NULL)
		fprintf(stderr, "Can't bind to locale %s\n", lc_all);

@ @<parse...@>=
for (i = 1; i < argc; i++)
{
	char *arg = argv[i];

	if (arg[0] == '-') {
		switch (arg[1]) {
			case 's':bs = 1; break;
			case 'p': 
				i++;
				prefix = argv[i];
				break;
			case 'v': bv = 1; break;
			case 'f': bf = 1; break;
			case 'u': i++; 
				  fontuid = atoi(argv[i]);
				  break;
			case 'n': i++;
				  strcpy (fontname, argv[i]);
				  break;
			default: 
				  showhelp();
				  exit (1);
		}
	}
	else if (ttf_file == NULL) ttf_file = arg;
	else if (facename == NULL) facename = arg;
	else fputs(_("! Extra file names ignored\n"),stderr);
}

if (ttf_file == NULL || facename == NULL) {
	fprintf(stderr, _("You must specify a TTF file and a face name!\n"));
	showhelp();
	exit (1);
}

sprintf(slface, "%ssl", facename);

if (fontname[0] != 0) strcpy(fam_name, fontname);

if (access(ttf_file, 0) == -1) 
	fatal (_("Can't open font file %s"), ttf_file);

ttf_basename = trimpath (ttf_file);


@ variable for store path names
@ @<variable...@>=
char dvipsdir[NAMELEN], type1dir[NAMELEN], tfmdir[NAMELEN], afmdir[NAMELEN];

@ make the directorties
@d MODE 0755
@<create dire...@>=
mkdir("fonts", MODE);
mkdir("fonts/afm", MODE);
mkdir("fonts/tfm", MODE);
mkdir("fonts/type1", MODE);
mkdir("fonts/afm/chinese", MODE);
mkdir("fonts/tfm/chinese", MODE);
mkdir("fonts/type1/chinese", MODE);

mkdir("dvips", MODE);
mkdir("dvips/chinese", MODE);

mkdir("pdftex", MODE);
mkdir("pdftex/config", MODE);

mkdir("dvipdfm", MODE);
mkdir("dvipdfm/config", MODE);

mkdir("tex", MODE);
mkdir("tex/latex", MODE);
mkdir("tex/latex/CJK", MODE);
mkdir("tex/latex/CJK/GB", MODE);

sprintf(dvipsdir,"dvips/chinese/%s%s/", prefix,facename);
mkdir(dvipsdir, MODE);

sprintf(afmdir,"fonts/afm/chinese/%s%s/", prefix,facename);
mkdir(afmdir, MODE);

sprintf(tfmdir,"fonts/tfm/chinese/%s%s/", prefix,facename);
mkdir(tfmdir, MODE);

sprintf(type1dir,"fonts/type1/chinese/%s%s/", prefix,facename);
mkdir(type1dir, MODE);

@ @<open map...@>=
if ((pfbmapfile = fopen ("cjk.map", "a")) == NULL) fatal(_("can't open cjk.map\n"));
sprintf(cmd,"pdftex/config/%s%s.map", prefix, facename);
if ((ttfmapfile = fopen (cmd, "w")) == NULL) fatal(_("can't open ttfonts.map\n"));
sprintf(cmd,"dvipdfm/config/%s%s.map", prefix, facename);
if ((dvipdfm_mapfile = fopen (cmd, "w")) == NULL) fatal(_("can't open cid-x.map\n"));

@ @<make the fon...@>=
strcpy(cmd, type1dir);
strcat(cmd, name);
strcat(cmd, ".ps");
strcpy(cmd1, dvipsdir);
strcat(cmd1, name);
strcat(cmd1, ".enc");

ttf2pfb(ttf_file, cmd, cmd1, nn, fontuid);

/* assemble row fonts into type1 pfb */
strcpy(cmd1, type1dir);
strcat(cmd1, name);
strcat(cmd1, ".pfb");
t1asm(cmd, cmd1);

/* delete the ps file */
unlink(cmd);

/* Make .afm from .pfb */
#ifdef INSERT
type1afm(cmd1);
#else
strcpy(cmd, "type1afm ");
strcat(cmd, cmd1);
system(cmd);
#endif


/* move afm file */
strcpy(cmd, type1dir);
strcat(cmd, name);
strcat(cmd, ".afm");
sprintf(cmd1, "%s%s.afm", afmdir, name);

#ifndef WIN32
if (link(cmd, cmd1) < 0) 
	if(errno == EEXIST) {
		unlink(cmd1);
		link(cmd, cmd1);
	} else {
		fprintf(stderr, _("Warning: can't move afm file to destination\n"));
		goto dontdelete;
}

unlink(cmd);
dontdelete:
#else
MoveFileEx(cmd, cmd1, MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED);
#endif

/* Make .tfm from .afm */
#ifndef WIN32
sprintf(cmd,"%s%s.afm", afmdir, name);
sprintf(cmd1,"%s%s.tfm", tfmdir, name);
if (fork() == 0) {
	fclose(stdout);
	execlp("afm2tfm", "afm2tfm", cmd, cmd1, NULL);
}
#else
sprintf(cmd, "afm2tfm %s%s.afm %s%s.tfm", afmdir, name, tfmdir, name);
system(cmd);
#endif


/* make slant tfm font */
#ifndef WIN32
sprintf(cmd,"%s%s.afm", afmdir, name);
sprintf(cmd1,"%s%s.tfm", tfmdir, slname);
if (fork() == 0){ 
	fclose(stdout);
	execlp("afm2tfm", "afm2tfm", cmd, "-s", ".167", cmd1, NULL);
}
#else
sprintf(cmd, "afm2tfm %s%s.afm -s .167 %s%s.tfm", afmdir, name, tfmdir, slname);
system(cmd);
#endif

@ @<add line...@>=
/* add line to map file for dvips use ttf fonts */
fprintf(ttfmapfile,"%s <%s.enc <%s\n", name, name, ttf_basename);
fprintf(ttfmapfile,"%s <%s.enc <%s\n", slname, name, ttf_basename);
fprintf(dvipdfm_mapfile,"%s none %s\n", name, name);
fprintf(dvipdfm_mapfile,"%s none %s -s 0.167\n", slname, name);

/* map file for dvips use pfb fonts */
fprintf(pfbmapfile,"%s %s%s <%s.pfb\n", name, fontname, plane, name);

fprintf(pfbmapfile,"%s %s%s \" .167 SlantFont \" < %s.pfb\n", 
		slname, fontname, plane, name);

@ @<close map...@>=
	fclose(ttfmapfile);
	fclose(pfbmapfile);
	fclose(dvipdfm_mapfile);

